import contextlib
import sys
from typing import Optional, Tuple
import uuid

from click.exceptions import ClickException

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models import (
    CloudAnalyticsEvent,
    CloudAnalyticsEventCommandName,
    CloudAnalyticsEventError,
    CloudAnalyticsEventName,
    CreateAnalyticsEvent,
)
from anyscale.cloud import get_cloud_id_and_name


def get_organization_default_cloud(api_client: DefaultApi) -> Optional[str]:
    """Return default cloud name for organization if it exists and
        if user has correct permissions for it.

        Returns:
            Name of default cloud name for organization if it exists and
            if user has correct permissions for it.
        """
    user = api_client.get_user_info_api_v2_userinfo_get().result
    organization = user.organizations[0]  # Each user only has one org
    if organization.default_cloud_id:
        try:
            # Check permissions
            _, cloud_name = get_cloud_id_and_name(
                api_client, cloud_id=organization.default_cloud_id
            )
            return str(cloud_name)
        except Exception:  # noqa: BLE001
            return None
    return None


def get_default_cloud(
    api_client: DefaultApi, cloud_name: Optional[str]
) -> Tuple[str, str]:
    """Returns the cloud id from cloud name.
    If cloud name is not provided, returns the default cloud name if exists in organization.
    If default cloud name does not exist returns last used cloud.
    """

    if cloud_name is None:
        default_cloud_name = get_organization_default_cloud(api_client)
        if default_cloud_name:
            cloud_name = default_cloud_name
    return get_cloud_id_and_name(api_client, cloud_name=cloud_name)


def verify_anyscale_access(
    api_client: DefaultApi, cloud_id: str, logger: BlockLogger
) -> bool:
    try:
        api_client.verify_access_api_v2_cloudsverify_access_cloud_id_get(cloud_id)
        return True
    except ClickException as e:
        logger.error(
            f"Anyscale's control plane is unable to access resources on your cloud provider.\n{e}"
        )
        return False


class CloudEventProducer:
    """
    Produce events during cloud setup/register/verify
    """

    def __init__(self, api_client: DefaultApi):
        self.api_client = api_client
        self.cloud_id: Optional[str] = None

    def init_trace_context(
        self,
        command_name: CloudAnalyticsEventCommandName,
        cloud_id: Optional[str] = None,
    ):
        self.trace_id = str(uuid.uuid4().hex)
        self.command_name = command_name
        self.raw_command_input = str(" ".join(sys.argv[1:]))
        self.cloud_id = cloud_id

    def set_cloud_id(self, cloud_id: str):
        self.cloud_id = cloud_id

    def produce(
        self,
        event_name: CloudAnalyticsEventName,
        succeeded: bool,
        internal_error: Optional[str] = None,
    ):
        with contextlib.suppress(Exception):
            # shouldn't block cloud setup even if cloud event generation fails
            error = None
            if not succeeded:
                # TODO (congding): cloud provider error
                error_message = internal_error if internal_error else ""
                error = CloudAnalyticsEventError(internal_error=error_message,)

            self.api_client.produce_analytics_event_api_v2_analytics_post(
                CreateAnalyticsEvent(
                    cloud_analytics_event=CloudAnalyticsEvent(
                        trace_id=self.trace_id,
                        cloud_id=self.cloud_id,
                        succeeded=succeeded,
                        command_name=self.command_name,
                        raw_command_input=self.raw_command_input,
                        event_name=event_name,
                        error=error,
                    )
                )
            )



from typing import Optional

from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from anyscale.sdk.anyscale_client.models.cloud import Cloud
from anyscale.sdk.anyscale_client.models.cluster_management_stack_versions import (
    ClusterManagementStackVersions,
)


def _is_v2_cloud(base_api: BaseApi, cloud_id: str) -> bool:
    cloud: Cloud = base_api.get_cloud(cloud_id).result
    cloud_version: Optional[
        ClusterManagementStackVersions
    ] = cloud.cluster_management_stack_version
    return not cloud_version or cloud_version == ClusterManagementStackVersions.V2



from functools import lru_cache
from typing import Optional

from anyscale.api_utils.cloud_util import _is_v2_cloud
from anyscale.api_utils.exceptions.job_errors import NoJobRunError
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from anyscale.sdk.anyscale_client.models.cluster_compute import ClusterCompute
from anyscale.sdk.anyscale_client.models.production_job import ProductionJob


def _get_job_run_id(
    base_api: BaseApi,
    *,
    job_id: Optional[str] = None,
    job_run_id: Optional[str] = None,
) -> str:
    assert bool(job_id) != bool(
        job_run_id
    ), "Exactly one of `job_id` or `job_run_id` must be provided."
    if job_id:
        prod_job: ProductionJob = base_api.get_production_job(job_id).result
        if not prod_job.last_job_run_id:
            raise NoJobRunError(f"Production job {job_id} has no job runs.")
        job_run_id = prod_job.last_job_run_id
    return job_run_id  # type: ignore


@lru_cache()
def _is_production_job_on_v2_cloud(base_api: BaseApi, production_job_id: str) -> bool:
    prod_job: ProductionJob = base_api.get_production_job(production_job_id).result
    cluster_compute: ClusterCompute = base_api.get_cluster_compute(
        prod_job.config.compute_config_id
    ).result
    return _is_v2_cloud(base_api, cluster_compute.config.cloud_id)



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class CreateBYODClusterEnvironmentConfigurationSchema(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'docker_image': 'str',
        'ray_version': 'str',
        'env_vars': 'object',
        'registry_login_secret': 'str'
    }

    attribute_map = {
        'docker_image': 'docker_image',
        'ray_version': 'ray_version',
        'env_vars': 'env_vars',
        'registry_login_secret': 'registry_login_secret'
    }

    def __init__(self, docker_image=None, ray_version=None, env_vars=None, registry_login_secret=None, local_vars_configuration=None):  # noqa: E501
        """CreateBYODClusterEnvironmentConfigurationSchema - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._docker_image = None
        self._ray_version = None
        self._env_vars = None
        self._registry_login_secret = None
        self.discriminator = None

        self.docker_image = docker_image
        self.ray_version = ray_version
        if env_vars is not None:
            self.env_vars = env_vars
        if registry_login_secret is not None:
            self.registry_login_secret = registry_login_secret

    @property
    def docker_image(self):
        """Gets the docker_image of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501

        The custom docker base image used in the Cluster Environment.  # noqa: E501

        :return: The docker_image of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501
        :rtype: str
        """
        return self._docker_image

    @docker_image.setter
    def docker_image(self, docker_image):
        """Sets the docker_image of this CreateBYODClusterEnvironmentConfigurationSchema.

        The custom docker base image used in the Cluster Environment.  # noqa: E501

        :param docker_image: The docker_image of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and docker_image is None:  # noqa: E501
            raise ValueError("Invalid value for `docker_image`, must not be `None`")  # noqa: E501

        self._docker_image = docker_image

    @property
    def ray_version(self):
        """Gets the ray_version of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501

        The version of Ray used in the customer docker image.  # noqa: E501

        :return: The ray_version of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501
        :rtype: str
        """
        return self._ray_version

    @ray_version.setter
    def ray_version(self, ray_version):
        """Sets the ray_version of this CreateBYODClusterEnvironmentConfigurationSchema.

        The version of Ray used in the customer docker image.  # noqa: E501

        :param ray_version: The ray_version of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and ray_version is None:  # noqa: E501
            raise ValueError("Invalid value for `ray_version`, must not be `None`")  # noqa: E501

        self._ray_version = ray_version

    @property
    def env_vars(self):
        """Gets the env_vars of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501

        Environment variables in the docker image that will be used at runtime.  # noqa: E501

        :return: The env_vars of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501
        :rtype: object
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this CreateBYODClusterEnvironmentConfigurationSchema.

        Environment variables in the docker image that will be used at runtime.  # noqa: E501

        :param env_vars: The env_vars of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501
        :type: object
        """

        self._env_vars = env_vars

    @property
    def registry_login_secret(self):
        """Gets the registry_login_secret of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501

        The name or identifier of a secret containing credentials to authenticate to the docker registry hosting the image.  # noqa: E501

        :return: The registry_login_secret of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501
        :rtype: str
        """
        return self._registry_login_secret

    @registry_login_secret.setter
    def registry_login_secret(self, registry_login_secret):
        """Sets the registry_login_secret of this CreateBYODClusterEnvironmentConfigurationSchema.

        The name or identifier of a secret containing credentials to authenticate to the docker registry hosting the image.  # noqa: E501

        :param registry_login_secret: The registry_login_secret of this CreateBYODClusterEnvironmentConfigurationSchema.  # noqa: E501
        :type: str
        """

        self._registry_login_secret = registry_login_secret

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateBYODClusterEnvironmentConfigurationSchema):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateBYODClusterEnvironmentConfigurationSchema):
            return True

        return self.to_dict() != other.to_dict()



import os
import shlex
import subprocess
from typing import Any, List, Optional

import tabulate

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.experimental_workspace import (
    ExperimentalWorkspace,
)
from anyscale.controllers.base_controller import BaseController
from anyscale.util import get_endpoint
from anyscale.utils.workspace_utils import (
    extract_workspace_parameters_from_cluster_config,
    WorkspaceCommandContext,
)
from anyscale.workspace import load_workspace_id_or_throw, write_workspace_id_to_disk


ANYSCALE_WORKSPACES_SSH_OPTIONS = [
    "-o",
    "StrictHostKeyChecking=no",
    "-o",
    "UserKnownHostsFile=/dev/null",
    "-o",
    "IdentitiesOnly=yes",
]


class WorkspaceController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log

    def list(self) -> None:
        """
        prints a non-exhaustive tabular list of information about non-deleted workspaces.
        """
        workspaces_data: List[
            ExperimentalWorkspace
        ] = self.api_client.list_workspaces_api_v2_experimental_workspaces_get().results

        workspaces_table: List[List[Any]] = [
            [
                workspace.name,
                workspace.id,
                workspace.state,
                get_endpoint(f"/workspaces/{workspace.id}/{workspace.cluster_id}"),
            ]
            for workspace in workspaces_data
            if not workspace.is_deleted
        ]

        table = tabulate.tabulate(
            workspaces_table, headers=["NAME", "ID", "STATE", "URL",], tablefmt="plain",
        )

        print(f"Workspaces:\n{table}")

    def clone(self, workspace: ExperimentalWorkspace) -> None:
        dir_name = workspace.name
        os.makedirs(dir_name)

        workspace_id = workspace.id
        write_workspace_id_to_disk(workspace_id, dir_name)

    def get_activated_workspace(self) -> ExperimentalWorkspace:
        workspace_id = load_workspace_id_or_throw()
        return self.api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get(
            workspace_id
        ).result

    def get_workspace_dir_name(self) -> str:
        """
        Currently the directory name of Workspaces/Clusters
        is determiend by the project name.
        """
        workspace = self.get_activated_workspace()
        project = self.api_client.get_project_api_v2_projects_project_id_get(
            workspace.project_id
        ).result
        return project.name

    def _get_workspace_command_context(self) -> WorkspaceCommandContext:
        workspace = self.get_activated_workspace()

        cluster_id = workspace.cluster_id
        project_id = workspace.project_id

        return extract_workspace_parameters_from_cluster_config(
            cluster_id,
            project_id,
            api_client=self.api_client,
            ssh_dir="/tmp/anyscale/.ssh",
        )

    def run_cmd(self, cmd: str, args: List[str]):
        workspace_command_context = self._get_workspace_command_context()

        ssh_info = workspace_command_context.ssh

        head_ip = workspace_command_context.head_ip
        target_host = f"{ssh_info.user}@{head_ip}"

        command = shlex.quote(cmd)

        ssh_command = (
            ["ssh"]
            + ANYSCALE_WORKSPACES_SSH_OPTIONS
            + [
                "-o",
                f"ProxyCommand=ssh -W %h:%p -i {ssh_info.key_path} -o StrictHostKeyChecking=accept-new {target_host}",
            ]
            + ["-tt", "-i", ssh_info.key_path]
            + ["ray@0.0.0.0", "-p", "5020"]
            + (args if args and len(args) > 0 else [""])
            + [f"bash -i -c {command}"]
        )

        subprocess.run(ssh_command)

    def run_rsync(
        self,
        *,
        local_path: str,
        remote_path: Optional[str] = None,
        down: bool,
        rsync_filters: List[str],
        rsync_excludes: List[str],
    ) -> None:
        """
        Args:
            local_path: The local path to sync from/to.
            remote_path: The remote path to sync to/from. If not provided, the
                remote path will be set to the working directory.
            down: Whether to sync from 1. remote to local, or 2. local to remote.
            rsync_filters: A list of rsync filters to apply.
            rsync_excludes: A list of rsync excludes to apply.
        """
        workspace_command_context = self._get_workspace_command_context()
        ssh_info = workspace_command_context.ssh
        head_ip = workspace_command_context.head_ip

        identity_file_option = ["-i", ssh_info.key_path]
        proxy_thorugh_ssh_user_option = [
            "-o",
            f"ProxyCommand=ssh -W %h:%p -i {ssh_info.key_path} -o StrictHostKeyChecking=accept-new {ssh_info.user}@{head_ip}",
        ]
        ray_user_ssh_port_option = ["-p", "5020"]
        base_ssh_command = (
            ["ssh"]
            + identity_file_option
            + ANYSCALE_WORKSPACES_SSH_OPTIONS
            + proxy_thorugh_ssh_user_option
            + ray_user_ssh_port_option
        )

        rsync_command = [
            "rsync",
            "--rsh",
            subprocess.list2cmdline(base_ssh_command),
            "-rvzl",
            "--delete",
        ]

        for rsync_exclude in rsync_excludes:
            rsync_command.extend(["--exclude", rsync_exclude])

        for rsync_filter in rsync_filters:
            rsync_command.extend(["--filter", f"dir-merge,- {rsync_filter}"])

        # TODO: Perhaps include this in cluster_config somewhere so that below isn't hardcoded
        # if either the user or host for the ray container is changed in the future
        ray_user_on_host = "ray@0.0.0.0"
        if not remote_path:
            # if not specified, default to the working directory
            remote_path = workspace_command_context.working_dir

        if down:
            rsync_command += [
                f"{ray_user_on_host}:{remote_path}",
                local_path,
            ]
        else:
            rsync_command += [
                local_path,
                f"{ray_user_on_host}:{remote_path}",
            ]

        subprocess.run(rsync_command)



import asyncio
import json
import os
from pathlib import Path
import tempfile
import time
from types import SimpleNamespace
from typing import Any, Dict, List, Optional, Tuple, Type
from unittest.mock import ANY, call, MagicMock, Mock, mock_open, patch

from aiohttp import ClientResponseError, ClientSession, WSMsgType
from asynctest import CoroutineMock
import click
from pydantic import ValidationError
import pytest
import yaml

from anyscale.api_utils.exceptions.log_retrieval_errors import (
    ExpectedLogRetrievalError,
    UnsupportedLogRetrievalMethodError,
)
from anyscale.api_utils.job_logs_util import _get_job_logs_from_storage_bucket
from anyscale.api_utils.job_util import _is_production_job_on_v2_cloud
from anyscale.api_utils.logs_util import _stream_log_from_ray_websocket
from anyscale.client.openapi_client.models.create_internal_production_job import (
    CreateInternalProductionJob,
)
from anyscale.client.openapi_client.models.ha_job_states import HaJobStates
from anyscale.client.openapi_client.models.production_job_config import (
    ProductionJobConfig,
)
from anyscale.controllers import job_controller as job_controller_module
from anyscale.controllers.job_controller import (
    JobController,
    LogProvider,
    LogsLogger,
    MiniJobRun,
)
from anyscale.models.job_model import _working_dir_is_remote_uri, JobConfig
from anyscale.sdk.anyscale_client.models.compute_template import ComputeTemplate
from anyscale.shared_anyscale_utils.test_util import (
    get_args_from_call,
    MockAsyncContextManagerReturnValue,
)
from anyscale.util import poll, PROJECT_NAME_ENV_VAR
from anyscale.utils.runtime_env import is_dir_remote_uri


CONDA_DICT = {"dependencies": ["pip", {"pip": ["pip-install-test==0.5"]}]}
PIP_LIST = ["requests==1.0.0", "pip-install-test"]
ENV_VARS_DICT = {"TEST_ENV_VAR": "test_value"}


class FakeLogger(LogsLogger):
    logs: List[str] = []

    def open_block(self, *args, **kwargs,) -> None:
        pass

    def log(self, msg: str):
        self.logs.append(msg)
        print(msg)


@pytest.fixture()
def test_directory():
    with tempfile.TemporaryDirectory() as tmp_dir:
        path = Path(tmp_dir)
        subdir = path / "subdir"
        subdir.mkdir(parents=True)
        requirements_file = subdir / "requirements.txt"
        with requirements_file.open(mode="w") as f:
            print("\n".join(PIP_LIST), file=f)

        good_conda_file = subdir / "good_conda_env.yaml"
        with good_conda_file.open(mode="w") as f:
            yaml.dump(CONDA_DICT, f)

        bad_conda_file = subdir / "bad_conda_env.yaml"
        with bad_conda_file.open(mode="w") as f:
            print("% this is not a YAML file %", file=f)

        old_dir = os.getcwd()
        os.chdir(tmp_dir)
        yield subdir, requirements_file, good_conda_file, bad_conda_file
        os.chdir(old_dir)


@pytest.fixture()
def patch_jobs_anyscale_api_client(base_mock_anyscale_api_client: Mock):
    base_mock_anyscale_api_client.get_cluster_environment_build = Mock(
        return_value=Mock(result=Mock(status="succeeded"))
    )
    with patch.multiple(
        "anyscale.cluster_env",
        get_auth_api_client=Mock(
            return_value=Mock(anyscale_api_client=base_mock_anyscale_api_client)
        ),
    ):
        yield


@pytest.mark.parametrize("workspace_id", [None, "test_workspace_id"])
def test_generate_config_from_entrypoint(
    mock_auth_api_client, workspace_id: Optional[str]
):
    mock_logger = Mock()
    job_controller = JobController(log=mock_logger)
    entrypoint = ["python", "test.py"]
    name = "test_name"
    description = "test_description"

    mock_get_default_cluster_compute = Mock(
        return_value=Mock(id="mock_compute_config_id")
    )

    mock_validate_successful_build = Mock()
    mock_get_default_cluster_env_build = Mock(return_value=Mock(id="mock_build_id"))
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
        get_default_cluster_compute=mock_get_default_cluster_compute,
        get_default_cluster_env_build=mock_get_default_cluster_env_build,
    ):
        job_config = job_controller.generate_config_from_entrypoint(
            entrypoint, name, description, workspace_id
        )

    assert job_config.entrypoint == "python test.py"
    assert job_config.name == name
    assert job_config.description == description
    assert job_config.workspace_id == workspace_id


@pytest.mark.parametrize("add_extra_field", [True, False])
def test_config_validation(add_extra_field, mock_auth_api_client):
    """
    This test checks that an error is thrown if extra/mispelled fields
    are added into the config file
    """
    mock_logger = Mock()
    job_controller = JobController(log=mock_logger)

    name = "test_job_name"
    description = "mock_description"
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    project_id = "test_project_id"
    extra_field_value = "*****"
    config_dict = {
        "name": name,
        "description": description,
        "entrypoint": "python test.py",
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
    }
    if add_extra_field:
        extra_field_value = "*****"
        config_dict["extra_field"] = extra_field_value

    mock_validate_successful_build = Mock()
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple("os.path", exists=Mock(return_value=True)), patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ), patch.multiple(
        "anyscale.controllers.job_controller", validate_job_config_dict=Mock(),
    ):
        if add_extra_field:
            with pytest.raises(ValidationError):
                job_controller.generate_config_from_file(
                    "test_job_config_file", name=name, description=description,
                )
        else:
            job_controller.generate_config_from_file(
                "test_job_config_file", name=name, description=description,
            )


@pytest.mark.parametrize(
    "config_dict",
    [
        {
            "entrypoint": "mock_entrypoint",
            "build_id": "mock_build_id",
            "compute_config_id": "cpt_123",
        },
        {
            "entrypoint": "mock_entrypoint",
            "cluster_env": "mock_cluster_env",
            "compute_config": "mock_compute_config",
        },
        {
            "entrypoint": "mock_entrypoint",
            "compute_config": {
                "cloud_id": "mock_cloud_id",
                "region": "mock_region",
                "head_node_type": {"name": "head", "instance_type": "m5.large"},
                "worker_node_types": [],
            },
        },
        {
            "entrypoint": "mock_entrypoint",
            "cluster_env": "mock_cluster_env",
            "cloud": "mock_cloud",
        },
        {"entrypoint": "mock_entrypoint", "cluster_env": "mock_cluster_env"},
        {"entrypoint": "mock_entrypoint", "cloud": "mock_cloud"},
        {"entrypoint": "mock_entrypoint"},
        {"entrypoint": "mock_entrypoint", "project_id": "specified_project_id"},
    ],
)
@pytest.mark.parametrize("use_default_project", [True, False])
def test_submit_job(
    mock_auth_api_client,
    config_dict: Dict[str, Any],
    use_default_project: bool,
    compute_template_test_data: ComputeTemplate,
) -> None:
    config_project_id = config_dict.get("project_id")
    mock_logger = Mock()
    job_controller = JobController(log=mock_logger)
    mock_project_definition = Mock()
    mock_project_definition.root = "/some/directory"

    def infer_project_id_mock(*args: Any, **kwargs: Any):
        if kwargs.get("project_id"):
            return kwargs.get("project_id")
        elif use_default_project:
            return "mock_default_project_id"
        else:
            return "mock_project_id"

    mock_infer_project_id = Mock(side_effect=infer_project_id_mock)
    mock_get_build_from_cluster_env_identifier = Mock(
        return_value=Mock(id="mock_build_id")
    )

    mock_override_runtime_env_config = Mock(return_value=None)

    job_controller.api_client.get_compute_template_api_v2_compute_templates_template_id_get = Mock(
        return_value=Mock(result=compute_template_test_data)
    )
    job_controller.api_client.search_compute_templates_api_v2_compute_templates_search_post = Mock(
        return_value=Mock(results=[compute_template_test_data])
    )
    mock_get_cluster_compute_from_name = Mock(return_value=compute_template_test_data)
    mock_get_default_cluster_compute = Mock(return_value=compute_template_test_data)
    mock_register_compute_template = Mock(return_value=compute_template_test_data)

    mock_validate_successful_build = Mock()
    mock_get_default_cluster_env_build = Mock(return_value=Mock(id="mock_build_id"))
    mock_get_parent_cloud_id_and_name_of_project = Mock(
        return_value=("mock_parent_cloud_id", "mock_parent_cloud_name")
    )
    with patch(
        "builtins.open", mock_open(read_data=json.dumps(config_dict))
    ), patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
        get_cluster_compute_from_name=mock_get_cluster_compute_from_name,
        get_default_cluster_compute=mock_get_default_cluster_compute,
        register_compute_template=mock_register_compute_template,
        get_build_from_cluster_env_identifier=mock_get_build_from_cluster_env_identifier,
        get_default_cluster_env_build=mock_get_default_cluster_env_build,
        get_parent_cloud_id_and_name_of_project=mock_get_parent_cloud_id_and_name_of_project,
    ), patch.multiple(
        "anyscale.controllers.job_controller",
        infer_project_id=mock_infer_project_id,
        override_runtime_env_config=mock_override_runtime_env_config,
    ), patch.object(
        JobController,
        "_get_maximum_uptime_output",
        return_value="mock maximum uptime output",
    ), patch.multiple(
        "os.path", exists=Mock(return_value=True)
    ):
        job_controller.submit(
            "mock_config_file", name="mock_name", description="mock_description"
        )
    mock_validate_successful_build.assert_called_once_with("mock_build_id")
    mock_infer_project_id.assert_called_once_with(
        job_controller.anyscale_api_client,
        job_controller.api_client,
        job_controller.log,
        project_id=config_dict.get("project_id"),
        cluster_compute_id=compute_template_test_data.id,
        cluster_compute=config_dict.get("compute_config"),
        cloud=config_dict.get("cloud"),
    )

    if config_project_id:
        final_project_id = config_project_id
    elif use_default_project:
        final_project_id = "mock_default_project_id"
    else:
        final_project_id = "mock_project_id"
    job_controller.api_client.create_job_api_v2_decorated_ha_jobs_create_post.assert_called_once_with(
        CreateInternalProductionJob(
            name="mock_name",
            description="mock_description",
            project_id=final_project_id,
            workspace_id=None,
            config=ProductionJobConfig(  # noqa: PIE804
                **{
                    "entrypoint": "mock_entrypoint",
                    "build_id": "mock_build_id",
                    "compute_config_id": compute_template_test_data.id,
                }
            ),
        )
    )
    assert mock_logger.info.call_count == 5
    if "cluster_env" not in config_dict and "build_id" not in config_dict:
        mock_get_default_cluster_env_build.assert_called_once_with()
    if "compute_config" in config_dict:
        compute_config = config_dict["compute_config"]
        if isinstance(compute_config, str):
            mock_get_cluster_compute_from_name.assert_called_with(compute_config)
        elif isinstance(compute_config, dict):
            mock_register_compute_template.assert_called_once_with(compute_config)
    elif "cloud" in config_dict:
        mock_get_default_cluster_compute.assert_called_once_with(
            cloud_name=config_dict["cloud"], project_id=None
        )
    elif "compute_config_id" in config_dict:
        mock_get_default_cluster_compute.assert_not_called()
    if "project_id" in config_dict:
        mock_get_parent_cloud_id_and_name_of_project.assert_called_once_with(
            config_dict.get("project_id")
        )


@pytest.mark.parametrize("include_all_users", [False, True])
@pytest.mark.parametrize("name", ["mock_job_name", None])
@pytest.mark.parametrize("job_id", ["mock_job_id", None])
@pytest.mark.parametrize("project_id", ["mock_project_id", None])
@pytest.mark.parametrize(
    "passed_service_id", [False, True]
)  # Whether `job_id` is id of job or service
@pytest.mark.parametrize("include_archived", [False, True])
def test_list_jobs(  # noqa: PLR0913
    mock_auth_api_client,
    include_all_users: bool,
    name: Optional[str],
    job_id: Optional[str],
    project_id: Optional[str],
    passed_service_id: bool,
    include_archived: bool,
) -> None:
    job_controller = JobController()
    job_controller.api_client.get_user_info_api_v2_userinfo_get = Mock(
        return_value=Mock(result=Mock(id="mock_user_id"))
    )
    job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get = Mock(
        return_value=Mock(
            results=[Mock(config=Mock(entrypoint=""))] * 10,
            metadata=Mock(next_paging_token="paging_token"),
        )
    )
    job_controller.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get = Mock(
        return_value=Mock(
            result=Mock(
                config=Mock(entrypoint="", is_service=passed_service_id),
                is_service=passed_service_id,
            )
        )
    )

    if passed_service_id and job_id is not None:
        # Raise error if trying to list id that is not valid for the command.
        # Eg: job_id providied for `anyscale service list`
        with pytest.raises(click.ClickException):
            job_controller.list(
                include_all_users=include_all_users,
                name=name,
                job_id=job_id,
                project_id=project_id,
                include_archived=include_archived,
                max_items=20,
            )
        return
    else:
        job_controller.list(
            include_all_users=include_all_users,
            name=name,
            job_id=job_id,
            project_id=project_id,
            include_archived=include_archived,
            max_items=20,
        )

    if job_id:
        job_controller.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get.assert_called_once_with(
            job_id
        )
        job_controller.api_client.get_user_info_api_v2_userinfo_get.assert_not_called()
        job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get.assert_not_called()
    else:
        creator_id: Optional[str] = None
        if not include_all_users:
            creator_id = "mock_user_id"
            job_controller.api_client.get_user_info_api_v2_userinfo_get.assert_called_once()
        else:
            job_controller.api_client.get_user_info_api_v2_userinfo_get.assert_not_called()
        job_controller.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get.assert_not_called()
        job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get.assert_any_call(
            creator_id=creator_id,
            name=name,
            project_id=project_id,
            type_filter="BATCH_JOB",
            archive_status="ALL" if include_archived else "NOT_ARCHIVED",
            count=10,
        )
        job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get.assert_any_call(
            creator_id=creator_id,
            name=name,
            project_id=project_id,
            type_filter="BATCH_JOB",
            archive_status="ALL" if include_archived else "NOT_ARCHIVED",
            count=10,
            paging_token="paging_token",
        )
        assert (
            job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get.call_count
            == 2
        )


@pytest.mark.parametrize("name", ["mock_job_name", None])
@pytest.mark.parametrize("id", ["mock_job_id", None])
def test_terminate_job(
    mock_auth_api_client, name: Optional[str], id: Optional[str],  # noqa: A002
) -> None:
    job_controller = JobController()
    job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get = Mock(
        return_value=Mock(results=[Mock(id="mock_job_id")])
    )
    job_controller.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get = Mock(
        return_value=Mock(result=Mock(id="mock_job_id"))
    )
    if not name and not id:
        with pytest.raises(click.ClickException):
            job_controller.terminate(id, name)
        return
    else:
        job_controller.terminate(id, name)

    job_controller.api_client.terminate_job_api_v2_decorated_ha_jobs_production_job_id_terminate_post.assert_called_once_with(
        "mock_job_id"
    )
    if name is not None and id is None:
        job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get.assert_called_once_with(
            name=name, type_filter="BATCH_JOB",
        )


@pytest.mark.parametrize("name", ["mock_job_name", None])
@pytest.mark.parametrize("id", ["mock_job_id", None])
def test_archive_job(
    mock_auth_api_client, name: Optional[str], id: Optional[str],  # noqa: A002
) -> None:
    job_controller = JobController()
    job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get = Mock(
        return_value=Mock(results=[Mock(id="mock_job_id")])
    )
    job_controller.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get = Mock(
        return_value=Mock(result=Mock(id="mock_job_id"))
    )
    if not name and not id:
        with pytest.raises(click.ClickException):
            job_controller.archive(id, name)
        return
    else:
        job_controller.archive(id, name)

    job_controller.api_client.archive_job_api_v2_decorated_ha_jobs_production_job_id_archive_post.assert_called_once_with(
        "mock_job_id"
    )
    if name is not None and id is None:
        job_controller.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get.assert_called_once_with(
            name=name, type_filter="BATCH_JOB",
        )


class TestValidateConda:
    def test_validate_conda_str(self, test_directory, patch_jobs_anyscale_api_client):
        """Tests the conda field is allowed to be a str. It represents an existing env name."""
        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"conda": "env_name", "working_dir": "s3://bucket"},
        )
        assert jc.runtime_env["conda"] == "env_name"

    def test_validate_conda_invalid_path(self, patch_jobs_anyscale_api_client):
        """If a path to a YAML file is given, it should error if the path does not exist."""
        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={"conda": "../bad_path.yaml"},
            )

    @pytest.mark.parametrize("absolute_path", [True, False])
    def test_validate_conda_valid_file(
        self, test_directory, absolute_path, patch_jobs_anyscale_api_client
    ):
        _, _, good_conda_file, _ = test_directory

        if absolute_path:
            good_conda_file = good_conda_file.resolve()

        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"conda": str(good_conda_file)},
        )
        assert jc.runtime_env["conda"] == CONDA_DICT

    @pytest.mark.parametrize("absolute_path", [True, False])
    def test_validate_conda_invalid_file(
        self, test_directory, absolute_path, patch_jobs_anyscale_api_client
    ):
        """We should error if a .yml file with invalid YAML format is specified."""
        _, _, _, bad_conda_file = test_directory

        if absolute_path:
            bad_conda_file = bad_conda_file.resolve()

        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={"conda": str(bad_conda_file)},
            )

    def test_validate_conda_valid_dict(self, patch_jobs_anyscale_api_client):
        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"conda": CONDA_DICT},
        )
        assert jc.runtime_env["conda"] == CONDA_DICT


class TestValidatePip:
    def test_validate_pip_invalid_path(self, patch_jobs_anyscale_api_client):
        """If a path to a .txt file is given, it should error if the path does not exist."""
        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={"pip": "../bad_path.txt"},
            )

    @pytest.mark.parametrize("absolute_path", [True, False])
    def test_validate_pip_valid_file(
        self, test_directory, absolute_path, patch_jobs_anyscale_api_client
    ):
        _, requirements_file, _, _ = test_directory

        if absolute_path:
            requirements_file = requirements_file.resolve()

        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"pip": str(requirements_file)},
        )
        assert jc.runtime_env["pip"] == PIP_LIST

    def test_validate_pip_valid_list(self, patch_jobs_anyscale_api_client):
        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"pip": PIP_LIST},
        )
        assert jc.runtime_env["pip"] == PIP_LIST


class TestValidateEnvVars:
    def test_validate_env_vars_valid_dict(self, patch_jobs_anyscale_api_client):
        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"env_vars": ENV_VARS_DICT},
        )
        assert jc.runtime_env["env_vars"] == ENV_VARS_DICT

    def test_validate_env_vars_invalid_dict(self, patch_jobs_anyscale_api_client):
        """Error if env_vars is not a dict."""
        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={"env_vars": "not_a_dict"},
            )

    def test_validate_env_vars_not_dict_str(self, patch_jobs_anyscale_api_client):
        """Error if env_vars is not a Dict[str, str]."""
        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={"env_vars": {"key", 123}},
            )
        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={"env_vars": {"key": "value", 123: "value2"}},
            )


class TestValidateUploadPath:
    @pytest.mark.parametrize("working_dir", [None, ".", "s3://bucket", "gs://bucket"])
    @pytest.mark.parametrize("upload_path", [None, "s3://bucket", "gs://bucket"])
    def test_validate_working_dir_and_upload_path_combinations(
        self, patch_jobs_anyscale_api_client, working_dir: str, upload_path: str
    ):
        """Test combinations of working_dir and upload_path
        Exceptions:
        1. Working_dir is a remote uri and upload_path is defined
        2. Working_dir is None and upload_path is defined
        """
        runtime_env = {}
        if working_dir:
            runtime_env["working_dir"] = working_dir
        if upload_path:
            runtime_env["upload_path"] = upload_path

        if is_dir_remote_uri(working_dir) and is_dir_remote_uri(upload_path):
            with pytest.raises(
                click.ClickException,
                match=r"`upload_path` was specified, but `working_dir` is not a local directory",
            ):
                JobConfig(
                    entrypoint="ls",
                    build_id="123",
                    compute_config_id="test",
                    runtime_env=runtime_env,
                )
        elif working_dir is None and is_dir_remote_uri(upload_path):
            with pytest.raises(
                click.ClickException,
                match=r"upload_path` was specified, but no `working_dir` is defined",
            ):
                JobConfig(
                    entrypoint="ls",
                    build_id="123",
                    compute_config_id="test",
                    runtime_env=runtime_env,
                )
        else:
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env=runtime_env,
            )

    def test_reject_non_s3_or_gs_upload_path(self, patch_jobs_anyscale_api_client):
        """If upload_path is specified, it must be an s3 or gs path."""
        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={"upload_path": "file://some_file"},
            )

    def test_accept_s3_upload_path(self, patch_jobs_anyscale_api_client):
        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"working_dir": "./", "upload_path": "s3://some_bucket"},
        )
        assert jc.runtime_env["upload_path"] == "s3://some_bucket"

    def test_accept_gs_upload_path(self, patch_jobs_anyscale_api_client):
        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"working_dir": "./", "upload_path": "gs://some_bucket"},
        )
        assert jc.runtime_env["upload_path"] == "gs://some_bucket"


class TestValidateWorkingDir:
    def test_working_dir_is_remote_uri(self):
        assert _working_dir_is_remote_uri("s3://some_bucket")
        assert _working_dir_is_remote_uri("s3://some_bucket/some/path")
        assert _working_dir_is_remote_uri("fake://some_bucket/some/path/")
        assert not _working_dir_is_remote_uri("/some/path")
        assert not _working_dir_is_remote_uri("some/path/")

    def test_reject_nonexistent_local_dir(self, patch_jobs_anyscale_api_client):
        """If a local working_dir is specified, it must exist."""
        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={
                    "working_dir": "/does/not/exist",
                    "upload_path": "gs://fake",
                },
            )

    def test_accept_uri(self, patch_jobs_anyscale_api_client):
        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"working_dir": "s3://path/to/archive.zip"},
        )
        assert jc.runtime_env["working_dir"] == "s3://path/to/archive.zip"


class TestValidatePyModules:
    def test_reject_local_dir(self, patch_jobs_anyscale_api_client):
        """Local directories are only supported using working_dir, not py_modules."""
        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={"py_modules": ["."]},
            )

        with pytest.raises(click.ClickException):
            JobConfig(
                entrypoint="ls",
                build_id="123",
                compute_config_id="test",
                runtime_env={"py_modules": ["/tmp/dir"]},
            )

    def test_accept_uri(self, patch_jobs_anyscale_api_client):
        jc = JobConfig(
            entrypoint="ls",
            build_id="123",
            compute_config_id="test",
            runtime_env={"py_modules": ["s3://path/to/archive.zip"]},
        )
        assert jc.runtime_env["py_modules"] == ["s3://path/to/archive.zip"]


class TestComputeConfig:
    @pytest.fixture()
    def patch_calls_to_get_compute_config(self,):
        mock_get_default_cluster_compute = Mock(
            return_value=Mock(id="mock_default_compute_config_id")
        )
        mock_get_cluster_compute_from_name = Mock(
            return_value=Mock(id="mock_existing_compute_config_id")
        )
        mock_register_compute_template = Mock(
            return_value=Mock(id="mock_registered_compute_config_id")
        )
        mock_get_parent_cloud_id_and_name_of_project = Mock(
            return_value=("mock_parent_cloud_id", "mock_parent_cloud_name")
        )

        with patch.multiple(
            "anyscale.models.job_model",
            get_default_cluster_compute=mock_get_default_cluster_compute,
            get_cluster_compute_from_name=mock_get_cluster_compute_from_name,
            register_compute_template=mock_register_compute_template,
            get_parent_cloud_id_and_name_of_project=mock_get_parent_cloud_id_and_name_of_project,
        ):
            yield mock_get_default_cluster_compute, mock_get_cluster_compute_from_name, mock_register_compute_template, mock_get_parent_cloud_id_and_name_of_project

    def test_fill_compute_config_id_from_compute_config_id(
        self, patch_jobs_anyscale_api_client, patch_calls_to_get_compute_config
    ):
        (
            mock_get_default_cluster_compute,
            mock_get_cluster_compute_from_name,
            mock_register_compute_template,
            mock_get_parent_cloud_id_and_name_of_project,
        ) = patch_calls_to_get_compute_config
        jc = JobConfig(
            entrypoint="ls", compute_config_id="mock_existing_compute_config_id",
        )
        assert jc.compute_config_id == "mock_existing_compute_config_id"
        mock_get_default_cluster_compute.assert_not_called()
        mock_get_cluster_compute_from_name.assert_not_called()
        mock_register_compute_template.assert_not_called()
        mock_get_parent_cloud_id_and_name_of_project.assert_not_called()

    def test_fill_compute_config_id_from_compute_config_name(
        self, patch_jobs_anyscale_api_client, patch_calls_to_get_compute_config
    ):
        (
            mock_get_default_cluster_compute,
            mock_get_cluster_compute_from_name,
            mock_register_compute_template,
            mock_get_parent_cloud_id_and_name_of_project,
        ) = patch_calls_to_get_compute_config
        jc = JobConfig(
            entrypoint="ls", compute_config="mock_existing_compute_config_name",
        )
        assert jc.compute_config_id == "mock_existing_compute_config_id"
        mock_get_default_cluster_compute.assert_not_called()
        mock_get_cluster_compute_from_name.assert_called_once_with(
            "mock_existing_compute_config_name"
        )
        mock_register_compute_template.assert_not_called()
        mock_get_parent_cloud_id_and_name_of_project.assert_not_called()

    def test_fill_compute_config_id_from_compute_config_dict(
        self, patch_jobs_anyscale_api_client, patch_calls_to_get_compute_config
    ):
        (
            mock_get_default_cluster_compute,
            mock_get_cluster_compute_from_name,
            mock_register_compute_template,
            mock_get_parent_cloud_id_and_name_of_project,
        ) = patch_calls_to_get_compute_config
        compute_config_dict = {
            "cloud_id": "mock_cloud_id",
            "region": "mock_region",
            "head_node_type": {"name": "head", "instance_type": "m5.large"},
            "worker_node_types": [],
        }
        jc = JobConfig(entrypoint="ls", compute_config=compute_config_dict,)
        assert jc.compute_config_id == "mock_registered_compute_config_id"
        mock_get_default_cluster_compute.assert_not_called()
        mock_get_cluster_compute_from_name.assert_not_called()
        mock_register_compute_template.assert_called_once_with(compute_config_dict)
        mock_get_parent_cloud_id_and_name_of_project.assert_not_called()

    def test_fill_compute_config_id_from_cloud(
        self, patch_jobs_anyscale_api_client, patch_calls_to_get_compute_config
    ):
        (
            mock_get_default_cluster_compute,
            mock_get_cluster_compute_from_name,
            mock_register_compute_template,
            mock_get_parent_cloud_id_and_name_of_project,
        ) = patch_calls_to_get_compute_config
        jc = JobConfig(entrypoint="ls", cloud="mock_cloud",)
        assert jc.compute_config_id == "mock_default_compute_config_id"
        mock_get_default_cluster_compute.assert_called_once_with(
            cloud_name="mock_cloud", project_id=None
        )
        mock_get_cluster_compute_from_name.assert_not_called()
        mock_register_compute_template.assert_not_called()
        mock_get_parent_cloud_id_and_name_of_project.assert_not_called()

    def test_fill_compute_config_id_from_project_with_cloud_isolation(
        self, patch_jobs_anyscale_api_client, patch_calls_to_get_compute_config
    ):
        (
            mock_get_default_cluster_compute,
            mock_get_cluster_compute_from_name,
            mock_register_compute_template,
            mock_get_parent_cloud_id_and_name_of_project,
        ) = patch_calls_to_get_compute_config
        jc = JobConfig(entrypoint="ls", project_id="mock_project_id")
        assert jc.compute_config_id == "mock_default_compute_config_id"
        mock_get_default_cluster_compute.assert_called_once_with(
            cloud_name="mock_parent_cloud_name", project_id=None
        )
        mock_get_cluster_compute_from_name.assert_not_called()
        mock_register_compute_template.assert_not_called()
        mock_get_parent_cloud_id_and_name_of_project.assert_called_once_with(
            "mock_project_id"
        )


@pytest.mark.parametrize("project_id", [None, "proj_id"])
@pytest.mark.parametrize("project_name", [None, "proj_name"])
@pytest.mark.parametrize("project_name_env_var", [None, "proj_name_env"])
def test_validate_project_id_field(
    project_id: Optional[str],
    project_name: Optional[str],
    project_name_env_var: Optional[str],
):
    mock_get_proj_id_from_name = Mock(return_value="proj_id")
    mock_validate_successful_build = Mock()
    config_dict = {
        "entrypoint": "mock_entrypoint",
        "build_id": "mock_build_id",
        "compute_config_id": "mock_compute_config_id",
        "project_id": project_id,
        "project": project_name,
    }
    mock_os_dict = (
        {PROJECT_NAME_ENV_VAR: project_name_env_var} if project_name_env_var else {}
    )
    with patch.multiple(
        "anyscale.models.job_model",
        get_proj_id_from_name=mock_get_proj_id_from_name,
        validate_successful_build=mock_validate_successful_build,
    ), patch.dict(os.environ, mock_os_dict):
        if project_id and project_name:
            with pytest.raises(click.ClickException):
                job_config = JobConfig.parse_obj(config_dict)
        else:
            job_config = JobConfig.parse_obj(config_dict)
            if project_name_env_var:
                assert job_config.project_id == "proj_id"
                mock_get_proj_id_from_name.assert_called_once_with(project_name_env_var)
            elif project_name:
                assert job_config.project_id == "proj_id"
                mock_get_proj_id_from_name.assert_called_once_with(project_name)
            else:
                assert job_config.project_id == project_id


def fake_log_provider(mode: str) -> Type[LogProvider]:
    class FakeLogProvider(LogProvider):
        def __init__(self, *args, **kwargs) -> None:
            self.constructed_at = int(time.time() * 1e9)
            self.page = -1
            self.query_count = 0

        def close(self):
            pass

        def query(  # noqa: PLR0911
            self, start_timestamp_ns: int, end_timestamp_ns: int
        ) -> List[Tuple[int, str]]:
            # For typing
            empty_list: List[Tuple[int, str]] = []

            if mode == "empty":
                return empty_list
            if mode == "single":
                if start_timestamp_ns > self.constructed_at:
                    return []
                return [(start_timestamp_ns, "start"), (self.constructed_at, "end")]
            if mode == "paginate":
                self.page += 1
                return [
                    [(start_timestamp_ns, "entry-1")],
                    [(start_timestamp_ns + 1, "entry-2")],
                    empty_list,
                ][self.page]
            if mode == "follow":
                self.page += 1
                return [
                    # simulate no log
                    empty_list,
                    empty_list,
                    [(start_timestamp_ns, "finally")],
                ][self.page]
            if mode == "flush_success":
                self.page += 1
                return (
                    [
                        # simulate no log
                        empty_list,
                        empty_list,
                        [(start_timestamp_ns, "finally")],
                        empty_list,
                        # Simulate 30 steps of empty logs
                    ]
                    + [empty_list for _ in range(30)]
                )[self.page]
            if mode == "wait_for_running":
                self.page += 1
                return (
                    [
                        # simulate no log
                        empty_list,
                        empty_list,
                        [(start_timestamp_ns, "finally")],
                        empty_list,
                        # Simulate 30 steps of empty logs
                    ]
                    + [empty_list for _ in range(30)]
                )[self.page]

            raise Exception("unreachable")

    return FakeLogProvider


def test_print_logs_for_active_job_run(mock_auth_api_client) -> None:
    job_controller = JobController(log=FakeLogger())
    job_id = "job_123"
    websocket_url = "http://localhost:5000/stream"
    job_controller.api_client = Mock(
        get_job_logs_stream_api_v2_logs_job_logs_stream_job_id_get=Mock(
            return_value=Mock(result=Mock(stream_url=websocket_url))
        )
    )
    mock_websocket = MockAsyncContextManagerReturnValue()

    # Test logs successfully streamed (followed) from cluster
    mock_websocket.receive = CoroutineMock(
        side_effect=[
            Mock(type=WSMsgType.TEXT, data=" abc "),
            Mock(type=WSMsgType.ERROR),
            Mock(type=WSMsgType.CLOSED),
        ]
    )
    with patch.object(
        ClientSession, ClientSession.ws_connect.__name__, return_value=mock_websocket
    ):
        job_controller._print_logs_for_active_job_run(job_id, should_follow=True)
    job_controller.api_client.get_job_logs_stream_api_v2_logs_job_logs_stream_job_id_get.assert_called_once_with(
        job_id=job_id
    )
    assert mock_websocket.receive.call_count == 3
    assert job_controller.log.logs == ["abc"]  # type: ignore

    # Test other exception thrown
    mock_other_exception = ClientResponseError(
        request_info=Mock(), history=Mock(), status=400
    )
    mock_websocket.receive = CoroutineMock(side_effect=mock_other_exception)
    with patch.object(
        ClientSession, ClientSession.ws_connect.__name__, return_value=mock_websocket
    ), pytest.raises(ClientResponseError):
        job_controller._print_logs_for_active_job_run(job_id, should_follow=True)

    # Test connection issues are raised and logged as error output
    # Indirectly tests `handle_and_log_lines_in_click_exception()`
    job_controller.log.error = Mock()  # type: ignore
    with patch.object(ClientSession, ClientSession.ws_connect.__name__), patch.object(
        job_controller_module,
        _stream_log_from_ray_websocket.__name__,
        new=CoroutineMock(side_effect=asyncio.TimeoutError),
    ):
        job_controller._print_logs_for_active_job_run(job_id, should_follow=True)
    log_content_args = job_controller.log.error.call_args_list[0]
    log_content: str = get_args_from_call(log_content_args)[0]
    assert (
        "only supported if the Anyscale client (SDK / CLI) can reach the cluster"
        in log_content
    )


def test_print_logs_for_inactive_job_run(mock_auth_api_client) -> None:
    job_id = "job_123"
    job_controller = JobController(log=Mock())
    job_controller.log.spinner = Mock(return_value=MockAsyncContextManagerReturnValue())  # type: ignore
    log_call_return_vals = ["", "logs\nretrieved"]

    # Test `_get_job_logs_from_storage_bucket`` called twice, and `log.log()` called twice with split logs
    with patch.object(
        job_controller_module, poll.__name__, return_value=["bucket", "hit", "twice"]
    ) as mock_poll, patch.object(
        job_controller_module,
        _get_job_logs_from_storage_bucket.__name__,
        new=CoroutineMock(side_effect=log_call_return_vals),
    ) as mock_get_job_logs_from_storage_bucket:
        job_controller._print_logs_for_inactive_job_run(job_id)
    job_controller.log.spinner.assert_called_once_with("Retrieving logs...")
    mock_poll.assert_called_once_with(interval_secs=5, max_iter=10)
    assert mock_get_job_logs_from_storage_bucket.call_count == len(log_call_return_vals)
    assert job_controller.log.log.call_count == 2  # type: ignore
    assert get_args_from_call(job_controller.log.log.call_args_list[0])[0] == "logs"  # type: ignore
    assert (
        get_args_from_call(job_controller.log.log.call_args_list[1])[0] == "retrieved"  # type: ignore
    )

    # Test `log.error()` called when logs are empty after polling
    empty_log_call_return_vals = ["", ""]
    with patch.object(
        job_controller_module, poll.__name__, return_value=["tried", "twice"]
    ) as mock_poll, patch.object(
        job_controller_module,
        _get_job_logs_from_storage_bucket.__name__,
        new=CoroutineMock(side_effect=empty_log_call_return_vals),
    ) as mock_get_job_logs_from_storage_bucket:
        job_controller._print_logs_for_inactive_job_run(job_id)
    assert mock_get_job_logs_from_storage_bucket.call_count == len(
        empty_log_call_return_vals
    )
    job_controller.log.error.assert_called_once_with(  # type: ignore
        "Logs are not available for this job run."
    )


@pytest.mark.parametrize(
    "mode",
    ["empty", "single", "paginate", "follow", "flush_success", "wait_for_running"],
)
def test_print_logs_legacy(mock_auth_api_client, capsys, mode) -> None:
    job_controller = JobController(log=FakeLogger())
    fake_log_provider_cls = fake_log_provider(mode)

    state = {"flush_success": "SUCCESS", "wait_for_running": "PENDING"}.get(
        mode, "RUNNING"
    )

    job_controller._get_formatted_latest_job_run = Mock(  # type: ignore
        return_value=MiniJobRun(
            last_job_run_id="mock_job_id", job_state=state, error=None
        )
    )

    def reset(*args, **kwargs):
        job_controller._get_formatted_latest_job_run = Mock(  # type: ignore
            return_value=MiniJobRun(
                last_job_run_id="mock_job_id", job_state="SUCCESS", error=None
            )
        )

    job_controller._wait_for_job_running = Mock(side_effect=reset)  # type: ignore
    job_controller.anyscale_api_client.get_job_logs_query_info = Mock(
        return_value=Mock(
            result=Mock(
                loki_dns_name="abc.com",
                access_token="secret-key",
                loki_query="query-abc",
            )
        )
    )

    def do_loki_logs() -> None:
        job_controller._print_logs_legacy(
            job=Mock(id="mock_job_id"),
            should_follow=mode in ["follow", "flush_success", "wait_for_running"],
            log_provider_cls=fake_log_provider_cls,
        )

    if mode == "empty":
        do_loki_logs()
        assert capsys.readouterr().out.strip() == ""
    elif mode == "single":
        do_loki_logs()
        assert capsys.readouterr().out.strip() == "start\nend"
    elif mode == "paginate":
        do_loki_logs()
        assert capsys.readouterr().out.strip() == "entry-1\nentry-2"
    elif mode == "follow":
        with pytest.raises(IndexError):  # controller should continue to query
            do_loki_logs()
        out = capsys.readouterr().out.strip()
        assert out == "finally"
    elif mode == "flush_success":
        do_loki_logs()
        out = capsys.readouterr().out.strip()
        assert out == "finally"
    elif mode == "wait_for_running":
        with pytest.raises(IndexError):  # controller should continue to query
            do_loki_logs()
        assert job_controller._wait_for_job_running.call_count == 1
        out = capsys.readouterr().out.strip()
        assert out == "finally"
    else:
        raise Exception("unreachable")


def test_print_logs_legacy_errors_when_job_on_v2_cloud(mock_auth_api_client) -> None:
    job_controller = JobController()
    mock_job = Mock(id="prodjob_123")
    with patch.object(
        job_controller_module,
        _is_production_job_on_v2_cloud.__name__,  # type: ignore
        return_value=True,
    ) as mock_is_production_job_on_v2_cloud, pytest.raises(
        UnsupportedLogRetrievalMethodError, match="Logs are not stored Anyscale-side"
    ):
        job_controller._print_logs_legacy(mock_job)
    mock_is_production_job_on_v2_cloud.assert_called_once_with(
        job_controller.anyscale_api_client, mock_job.id
    )


def test_print_logs_legacy_errors_when_not_logged_at_all(mock_auth_api_client) -> None:
    job_controller = JobController()
    mock_job = Mock(
        state=Mock(current_state=HaJobStates.TERMINATED, error=None),
        last_job_run_id="job_123",
    )
    fake_log_provider_cls = fake_log_provider("empty")
    with patch.object(
        job_controller_module,
        _is_production_job_on_v2_cloud.__name__,  # type: ignore
        return_value=False,
    ), pytest.raises(ExpectedLogRetrievalError):
        job_controller._print_logs_legacy(
            mock_job, log_provider_cls=fake_log_provider_cls
        )


def test_logs(mock_auth_api_client) -> None:
    job_controller = JobController()
    prod_job_id = "prodjob_123"
    last_job_run_id = "job_123"
    mock_job = Mock(
        id=prod_job_id,
        last_job_run_id=last_job_run_id,
        state=Mock(current_state=HaJobStates.TERMINATED),
    )
    job_controller._wait_for_a_job_run = Mock(return_value=mock_job)  # type: ignore
    job_controller._print_logs_for_active_job_run = mock_stream_logs = Mock()  # type: ignore
    job_controller._print_logs_legacy = mock_loki_logs = Mock()  # type: ignore
    job_controller._print_logs_for_inactive_job_run = mock_download_logs = Mock()  # type: ignore
    mock_ff_response = Mock()
    mock_ff_response.result = Mock()
    mock_ff_response.result.is_on = False
    job_controller.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get = Mock(
        return_value=mock_ff_response
    )

    def reset_mocks() -> None:
        for mock in [mock_stream_logs, mock_loki_logs, mock_download_logs]:
            mock.reset_mock()

    def do_assertions(
        called: Optional[List[Mock]] = None, not_called: Optional[List[Mock]] = None
    ) -> None:
        if not called:
            called = []
        if not not_called:
            not_called = []
        for mock in called:
            mock.assert_called_once()
        for mock in not_called:
            mock.assert_not_called()

    # Test Loki logs and does not follow by default if job is not in RUNNING state
    job_controller.logs(job_id=prod_job_id)
    do_assertions(not_called=[mock_stream_logs, mock_download_logs])
    mock_loki_logs.assert_called_once_with(mock_job, False)

    # Test download logs (ie. falls back) if Loki has no logs (raises `ExpectedLogRetrievalError`)
    reset_mocks()
    job_controller._print_logs_legacy.side_effect = ExpectedLogRetrievalError
    job_controller.logs(job_id=prod_job_id)
    do_assertions(
        called=[mock_loki_logs, mock_download_logs], not_called=[mock_stream_logs]
    )

    # Test stream if job is RUNNING
    reset_mocks()
    mock_job.state.current_state = HaJobStates.RUNNING
    job_controller.logs(job_id=prod_job_id)
    do_assertions(
        called=[mock_stream_logs], not_called=[mock_loki_logs, mock_download_logs]
    )

    # Test `all_attempts = True`
    reset_mocks()
    job_controller._get_all_job_runs = Mock(  # type: ignore
        return_value=[Mock(id="job_1"), Mock(id=last_job_run_id),]
    )
    job_controller.logs(job_id=prod_job_id, should_follow=True, all_attempts=True)
    job_controller._get_all_job_runs.assert_called_once_with(prod_job_id)
    mock_download_logs.assert_called_once_with("job_1")
    mock_stream_logs.assert_called_once_with(last_job_run_id, True)


def test_logs_streaming(mock_auth_api_client) -> None:
    job_controller = JobController()
    prod_job_id = "prodjob_123"
    last_job_run_id = "job_123"
    mock_job = Mock(
        id=prod_job_id,
        last_job_run_id=last_job_run_id,
        state=Mock(current_state=HaJobStates.RUNNING),
    )
    job_controller.log = Mock()
    job_controller.log.spinner = LogsLogger().spinner
    job_controller._wait_for_a_job_run = Mock(return_value=mock_job)  # type: ignore
    mock_ff_response = Mock()
    mock_ff_response.result = Mock()
    mock_ff_response.result.is_on = True
    job_controller.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get = Mock(
        return_value=mock_ff_response
    )

    get_jobs_logs_from_storage_bucket_streaming_mock = CoroutineMock(
        side_effect=[
            ("1\n2\n3\n", "next_token", 10, False),
            ("4\n5\n6\n", "next_token_2", 10, True),
            ("7\n8\n9\n10\n", "next_token_3", 10, True),
            ("", "next_token_3", 10, True),
        ]
    )

    with patch(
        "anyscale.controllers.job_controller._get_job_logs_from_storage_bucket_streaming",
        new=get_jobs_logs_from_storage_bucket_streaming_mock,
    ):
        job_controller.logs(job_id=prod_job_id, should_follow=True)

    get_jobs_logs_from_storage_bucket_streaming_mock.assert_has_awaits(
        [
            call(
                ANY,
                ANY,
                job_run_id=last_job_run_id,
                remove_escape_chars=False,
                next_page_token=None,
                cluster_journal_events_start_line=0,
            ),
            call(
                ANY,
                ANY,
                job_run_id=last_job_run_id,
                remove_escape_chars=False,
                next_page_token="next_token",
                cluster_journal_events_start_line=10,
            ),
            call(
                ANY,
                ANY,
                job_run_id=last_job_run_id,
                remove_escape_chars=False,
                next_page_token="next_token_2",
                cluster_journal_events_start_line=10,
            ),
            call(
                ANY,
                ANY,
                job_run_id=last_job_run_id,
                remove_escape_chars=False,
                next_page_token="next_token_3",
                cluster_journal_events_start_line=10,
            ),
        ]
    )
    job_controller.log.log.assert_has_calls(
        [
            call("1"),
            call("2"),
            call("3"),
            call("4"),
            call("5"),
            call("6"),
            call("7"),
            call("8"),
            call("9"),
            call("10"),
        ]
    )


def test_wait_for_job_running(mock_auth_api_client) -> None:
    job_controller = JobController(log=FakeLogger())
    i = 0

    def get_latest_job_run(*args, **kwargs):
        print("Called get_latest_job_run")
        nonlocal i
        i += 1
        ret = Mock()
        status = "RUNNING"
        if i < 3:
            status = "PENDING"
        elif i < 5:
            status = "AWAITING_CLUSTER_START"
        ret.state.current_state = status
        return ret

    job_controller._get_job = Mock(side_effect=get_latest_job_run)  # type: ignore
    job_controller._wait_for_job_running("fake_job_id")
    assert i == 5


def test_job_submit_parse_logic(mock_auth_api_client) -> None:
    job_controller = JobController()
    job_controller.generate_config_from_entrypoint = Mock()  # type: ignore
    job_controller.generate_config_from_file = Mock()  # type: ignore
    job_controller.submit_from_config = Mock()  # type: ignore

    # We are not in a workspace, so entrypoint should not be allowed
    with pytest.raises(click.ClickException):
        job_controller.submit(
            "file", entrypoint=["entrypoint"], is_entrypoint_cmd=False
        )

    with pytest.raises(click.ClickException):
        job_controller.submit("file", entrypoint=["entrypoint"], is_entrypoint_cmd=True)

    with pytest.raises(click.ClickException):
        job_controller.submit(
            "file", entrypoint=["entrypoint", "commands"], is_entrypoint_cmd=True
        )

    # Simulate a workspace
    with patch.dict(
        os.environ, {"ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": "fake_workspace_id"}
    ):
        # Fails due to is_entrypoint_cmd being False
        with pytest.raises(click.ClickException):
            job_controller.submit(
                "file", entrypoint=["entrypoint"], is_entrypoint_cmd=False
            )

        mock_config = Mock()
        job_controller.generate_config_from_file.return_value = mock_config
        job_controller.submit("file", entrypoint=[], is_entrypoint_cmd=False)
        job_controller.generate_config_from_file.assert_called_once_with(
            "file", name=None, description=None, workspace_id="fake_workspace_id"
        )
        job_controller.submit_from_config.assert_called_once_with(mock_config)
        job_controller.generate_config_from_file.reset_mock()
        job_controller.submit_from_config.reset_mock()

        mock_config = Mock()
        job_controller.generate_config_from_entrypoint.return_value = mock_config
        job_controller.submit("file", entrypoint=["entrypoint"], is_entrypoint_cmd=True)
        job_controller.generate_config_from_entrypoint.assert_called_once_with(
            ["file", "entrypoint"], None, None, "fake_workspace_id"
        )
        job_controller.submit_from_config.assert_called_once_with(mock_config)


def test_job_wait(mock_auth_api_client) -> None:
    job_controller = JobController()
    job_controller.api_client = Mock()
    job_controller._resolve_job_object = Mock(return_value=SimpleNamespace(id="id"))  # type: ignore
    spin_mock = MagicMock()
    job_controller.log = MagicMock()
    job_controller.log.spinner.return_value.__enter__.return_value = spin_mock

    def get_job_state(states_to_return, mock_state=None):
        if mock_state is None:
            mock_state = Mock()

        mock_state.i = 0

        def g(*args, **kwargs):
            nonlocal mock_state
            state = states_to_return[mock_state.i]
            mock_state.i += 1
            mock_state.state = state
            return state

        return g, mock_state

    # Test the happy path
    states_to_return = ["PENDING", "AWAITING_CLUSTER_START", "SUCCESS"]
    job_controller._get_job_state, mock_state = get_job_state(states_to_return)  # type: ignore

    assert job_controller.wait("name")
    assert mock_state.i == 3
    assert mock_state.state == "SUCCESS"
    spin_mock.succeed.assert_called_once()

    # Test the failure case
    spin_mock.reset_mock()
    states_to_return = [
        "PENDING",
        "AWAITING_CLUSTER_START",
        "ERRORED",
        "OUT_OF_RETRIES",
    ]
    job_controller._get_job_state, mock_state = get_job_state(states_to_return)  # type: ignore

    with pytest.raises(click.ClickException):
        job_controller.wait("name")
    assert mock_state.i == 4
    assert mock_state.state == "OUT_OF_RETRIES"
    spin_mock.fail.assert_called_once()

    # Test the non default case
    spin_mock.reset_mock()
    states_to_return = [
        "PENDING",
        "AWAITING_CLUSTER_START",
        "ERRORED",
        "SHOULD_NOT_REACH",
        "SHOULD_NOT_REACH",
        "SHOULD_NOT_REACH",
    ]
    job_controller._get_job_state, mock_state = get_job_state(states_to_return)  # type: ignore

    assert job_controller.wait("name", target_state="ERRORED")
    assert mock_state.state == "ERRORED"
    assert mock_state.i == 3
    spin_mock.succeed.assert_called_once()

    # Test timeout logic
    spin_mock.reset_mock()
    states_to_return = [
        "PENDING",
        "AWAITING_CLUSTER_START",
        "ERRORED",
        "SHOULD_NOT_REACH",
        "SHOULD_NOT_REACH",
        "SHOULD_NOT_REACH",
    ]
    job_controller._get_job_state, mock_state = get_job_state(states_to_return)  # type: ignore

    with pytest.raises(click.ClickException):
        job_controller.wait("name", timeout_secs=0.00001)

    # Test parsing logic
    with pytest.raises(click.ClickException):
        job_controller.wait("name", target_state="FAKE_STATE")


def test_override_runtime_env_config_and_infer_project_id(mock_auth_api_client):
    """Test helper functions in submit_from_config function
    Test that runtime_env is overwrriten from override_runtime_env_config
    Test the project_id is autopopulated from infer_project_id
    """
    config_dict = {
        "name": "mock_name",
        "entrypoint": "mock_entrypoint",
        "description": "mock_description",
        "build_id": "mock_build_id",
        "compute_config_id": "mock_compute_config_id",
    }

    assert "project_id" not in config_dict

    mock_project_id = "default"

    with patch.multiple(
        "anyscale.models.job_model", validate_successful_build=Mock(),
    ):
        job_config = JobConfig.parse_obj(config_dict)

    new_runtime_env = {"working_dir": "s3://bucket"}
    mock_override_config = Mock(return_value=new_runtime_env)

    mock_infer_project_id = Mock(return_value=mock_project_id)

    with patch.multiple(
        "anyscale.controllers.job_controller",
        infer_project_id=mock_infer_project_id,
        override_runtime_env_config=mock_override_config,
    ):
        job_controller = JobController()
        job_controller.api_client.create_job_api_v2_decorated_ha_jobs_create_post = (
            Mock()
        )
        job_controller._get_maximum_uptime_output = Mock()
        job_controller.submit_from_config(job_config)
        job_controller.api_client.create_job_api_v2_decorated_ha_jobs_create_post.assert_called_once_with(
            CreateInternalProductionJob(
                name=config_dict["name"],
                description=config_dict["description"],
                project_id=mock_project_id,
                config=ProductionJobConfig(
                    entrypoint=job_config.entrypoint,
                    runtime_env=new_runtime_env,
                    build_id=job_config.build_id,
                    compute_config_id=job_config.compute_config_id,
                    max_retries=job_config.max_retries,
                ),
            )
        )



import ast
import json
from unittest.mock import Mock, patch, PropertyMock

import click
import pytest

import anyscale
from anyscale.authenticate import (
    CREDENTIALS_FILE,
    CREDENTIALS_FILE_PERMISSIONS,
)
from anyscale.client.openapi_client.models.organization import Organization
from anyscale.client.openapi_client.models.user_info import UserInfo
from anyscale.controllers.auth_controller import AuthController


url = anyscale.util.get_endpoint("/credentials")


def test_auth_token_get_json_string():
    """
    Test auth_token_get_json_string function by turning a random token
    into a json-formatted string and then load it back
    """
    random_str = "This/IsA#}{RanDom..stRiN9"

    assert json.loads(AuthController()._auth_token_get_json_string(random_str)) == {
        "cli_token": random_str
    }


def test_auth_info_to_formatted_string():
    """
    Test if auth_info_to_formatted_string returns a string with all necessary info
    """
    userinfo = UserInfo(
        name="Anyscale Test",
        email="anyscale@anyscale.com",
        id="usr_somerandomid",
        organization_permission_level="collaborator",
        organizations=[
            Organization(
                name="Anyscale Inc.",
                id="org_someorgid1",
                local_vars_configuration=Mock(client_side_validation=False),
            ),
            Organization(
                name="Noscale Inc.",
                id="org_someorgid2",
                local_vars_configuration=Mock(client_side_validation=False),
            ),
        ],
        local_vars_configuration=Mock(client_side_validation=False),
    )
    s = AuthController()._auth_info_to_formatted_string(userinfo)

    assert (
        s == "  user name:                    Anyscale Test\n"
        "  user email:                   anyscale@anyscale.com\n"
        "  user id:                      usr_somerandomid\n"
        "  organization name:            Anyscale Inc.\n"
        "  organization id:              org_someorgid1\n"
        "  organization name:            Noscale Inc.\n"
        "  organization id:              org_someorgid2\n"
        "  organization role:            collaborator"
    )


def test_auth_info_to_formatted_string_cannot_parse():
    """
    Test auth_info_to_formatted_string when the UserInfo has been changed thus unparseable.
    In that case, the function should dump the entire UserInfo object in json format.
    We simulate it by raising TypeError when UserInfo.name is first referenced
    """
    userinfo = UserInfo(
        name="Anyscale Test",
        email="anyscale@anyscale.com",
        id="usr_somerandomid",
        organization_permission_level="collaborator",
        organizations=[
            Organization(
                name="Anyscale Inc.",
                id="org_someorgid1",
                local_vars_configuration=Mock(client_side_validation=False),
            ),
        ],
        local_vars_configuration=Mock(client_side_validation=False),
    )

    with patch(
        "anyscale.controllers.auth_controller.UserInfo.name", new_callable=PropertyMock
    ) as mock_name:
        # simulate TypeError on the first referrence
        mock_name.side_effect = [AttributeError(), None]
        s = AuthController()._auth_info_to_formatted_string(userinfo)

    userinfo = ast.literal_eval(s)
    assert userinfo["email"] == "anyscale@anyscale.com"
    assert userinfo["id"] == "usr_somerandomid"
    assert userinfo["organizations"][0]["name"] == "Anyscale Inc."
    assert userinfo["organizations"][0]["id"] == "org_someorgid1"
    assert userinfo["organization_permission_level"] == "collaborator"


def test_auth_set_first_time():
    """
    The first-time auth set command should succeed,
    and it also should not trigger AuthenticationBlock to be initialized.
    """
    mock_click_prompt = Mock(return_value="random_value_that_works   ")
    mock_click_confirm = Mock(return_value=True)
    mock_authentication_block_init = Mock(return_value=None)

    with patch.multiple("os.path", exists=Mock(return_value=False)), patch.multiple(
        "click", confirm=mock_click_confirm, prompt=mock_click_prompt,
    ), patch(
        "anyscale.controllers.auth_controller.credentials_check_sanity",
        Mock(return_value=True),
    ), patch(
        "anyscale.controllers.auth_controller.AuthController._save_credentials", Mock()
    ) as mock_save_credentials, patch(
        "anyscale.authenticate.AuthenticationBlock.__init__",
        mock_authentication_block_init,
    ):
        auth_controller = AuthController()
        auth_controller.set()

    # the value should be stripped
    mock_save_credentials.assert_called_once_with("random_value_that_works")
    mock_click_confirm.assert_not_called()
    mock_click_prompt.assert_called_once()
    mock_authentication_block_init.assert_not_called()


def test_auth_set_without_user_interaction():
    """
    Test when the user uses `set` function with the token provided.
    This should not trigger user interaction and must save the credentials file directly.
    """
    mock_click_prompt = Mock(return_value="token_via_prompt")
    mock_click_confirm = Mock(return_value=True)
    with patch.multiple("os.path", exists=Mock(return_value=False)), patch.multiple(
        "click", confirm=mock_click_confirm, prompt=mock_click_prompt,
    ), patch(
        "anyscale.controllers.auth_controller.credentials_check_sanity",
        Mock(return_value=True),
    ), patch(
        "anyscale.controllers.auth_controller.AuthController._save_credentials", Mock()
    ) as mock_save_credentials:
        auth_controller = AuthController()
        auth_controller.set("token_via_argument")

    mock_save_credentials.assert_called_once_with("token_via_argument")
    mock_click_confirm.assert_not_called()
    mock_click_prompt.assert_not_called()


def test_auth_set_without_user_interaction_invalid():
    """
    Test when the user uses `set` function with the token provided, but the token fails the sanity check.
    This should not trigger user interaction and also raise an exception.
    """
    mock_click_prompt = Mock(return_value="token_via_prompt")
    mock_click_confirm = Mock(return_value=True)
    with patch.multiple("os.path", exists=Mock(return_value=False)), patch.multiple(
        "click", confirm=mock_click_confirm, prompt=mock_click_prompt,
    ), patch(
        "anyscale.controllers.auth_controller.credentials_check_sanity",
        Mock(return_value=False),
    ), patch(
        "anyscale.controllers.auth_controller.AuthController._save_credentials", Mock()
    ) as mock_save_credentials, pytest.raises(
        click.ClickException
    ) as e:
        auth_controller = AuthController()
        auth_controller.set("token_via_argument_but_invalid")

    assert "invalid" in e.value.message.lower()
    mock_save_credentials.assert_not_called()
    mock_click_confirm.assert_not_called()
    mock_click_prompt.assert_not_called()


def test_auth_set_invalid():
    """
    Test auth set when user enters token, but sanity check fails.
    Should not save credentials, and must raise an exception.
    """
    with patch.multiple("os.path", exists=Mock(return_value=False)), patch.multiple(
        "click", prompt=Mock(return_value="incorrectcredential")
    ), patch(
        "anyscale.controllers.auth_controller.credentials_check_sanity",
        Mock(return_value=False),
    ) as mock_credentials_check_sanity, patch(
        "anyscale.controllers.auth_controller.AuthController._save_credentials", Mock()
    ) as mock_save_credentials, pytest.raises(
        click.ClickException
    ) as e:
        auth_controller = AuthController()
        auth_controller.set()

    assert "invalid" in e.value.message.lower()
    mock_credentials_check_sanity.assert_called_once_with("incorrectcredential")
    mock_save_credentials.assert_not_called()


def test_auth_set_file_exists():
    """
    Test auth set when the credential file exists
    and the user accepts to replace it.
    """
    mock_os_path_exists = Mock(return_value=True)
    mock_click_confirm = Mock(return_value=True)
    mock_click_prompt = Mock(return_value="correctcredential")

    with patch.multiple("os.path", exists=mock_os_path_exists), patch.multiple(
        "click", confirm=mock_click_confirm, prompt=mock_click_prompt
    ), patch(
        "anyscale.controllers.auth_controller.credentials_check_sanity",
        Mock(return_value=True),
    ), patch(
        "anyscale.controllers.auth_controller.AuthController._save_credentials", Mock()
    ) as mock_save_credentials:
        auth_controller = AuthController(path="/home/usr0/.anyscale/credentials.json",)
        auth_controller.set()

    mock_os_path_exists.assert_called_once_with("/home/usr0/.anyscale/credentials.json")
    mock_click_confirm.assert_called_once()
    mock_click_prompt.assert_called_once()
    mock_save_credentials.assert_called_once_with("correctcredential")


def test_auth_set_file_exists_cancel():
    """
    Test auth set when the credential file exists
    and the user denies to replace it
    """
    mock_os_path_exists = Mock(return_value=True)
    mock_click_confirm = Mock(return_value=False)

    with patch.multiple("os.path", exists=mock_os_path_exists), patch.multiple(
        "click", confirm=mock_click_confirm
    ), patch(
        "anyscale.controllers.auth_controller.AuthController._save_credentials", Mock()
    ) as mock_save_credentials:
        auth_controller = AuthController(path="/home/usr0/.anyscale/credentials.json")
        auth_controller.set()

    mock_os_path_exists.assert_called_once_with("/home/usr0/.anyscale/credentials.json")
    mock_click_confirm.assert_called_once()
    mock_save_credentials.assert_not_called()


def test_auth_fix_file_not_exist():
    """
    Test auth fix command when the file does not exist.
    Must raise an exception
    """
    mock_os_path_exists = Mock(return_value=False)
    mock_os_chmod = Mock()

    with patch("os.path.exists", mock_os_path_exists), patch(
        "os.chmod", mock_os_chmod
    ), pytest.raises(click.ClickException) as e:
        auth_controller = AuthController()
        auth_controller.fix()

    assert CREDENTIALS_FILE in e.value.message
    mock_os_chmod.assert_not_called()


def test_auth_fix():
    """
    Test auth fix command when the file exists.
    This should not trigger AuthenticationBlock to be initialized.
    """
    mock_os_path_exists = Mock(return_value=True)
    mock_os_chmod = Mock()
    mock_authentication_block_init = Mock(return_value=None)

    with patch("os.path.exists", mock_os_path_exists), patch(
        "os.chmod", mock_os_chmod
    ), patch(
        "anyscale.authenticate.AuthenticationBlock.__init__",
        mock_authentication_block_init,
    ):
        auth_controller = AuthController(path="/home/usr0/.anyscale/credentials.json")
        auth_controller.fix()

    mock_os_chmod.assert_called_once_with(
        "/home/usr0/.anyscale/credentials.json", CREDENTIALS_FILE_PERMISSIONS
    )
    mock_authentication_block_init.assert_not_called()


def test_auth_remove_file_not_exist():
    """
    Test auth remove command when the file does not exist.
    Must return without any error.
    """
    mock_os_path_exists = Mock(return_value=False)
    mock_click_confirm = Mock()
    mock_os_remove = Mock()

    with patch("os.path.exists", mock_os_path_exists), patch(
        "click.confirm", mock_click_confirm
    ), patch("os.remove", mock_os_remove), pytest.raises(click.ClickException) as e:
        auth_controller = AuthController()
        auth_controller.remove()

    mock_click_confirm.assert_not_called()
    mock_os_remove.assert_not_called()
    assert "file does not exist" in e.value.message


def test_auth_remove_user_abort():
    """
    Test auth remove when the file exists, but the user aborts.
    """
    mock_os_path_exists = Mock(return_value=True)
    mock_click_confirm = Mock(return_value=False)
    mock_os_remove = Mock()

    with patch("os.path.exists", mock_os_path_exists), patch(
        "click.confirm", mock_click_confirm
    ), patch("os.remove", mock_os_remove):
        auth_controller = AuthController()
        auth_controller.remove()

    mock_os_remove.assert_not_called()


def test_auth_remove():
    """
    Test auth remove when the file exists and the user accepts.
    This should not trigger AuthenticationBlock to be initialized.
    """
    mock_os_path_exists = Mock(return_value=True)
    mock_click_confirm = Mock(return_value=True)
    mock_os_remove = Mock()
    mock_authentication_block_init = Mock(return_value=None)

    with patch("os.path.exists", mock_os_path_exists), patch(
        "click.confirm", mock_click_confirm
    ), patch("os.remove", mock_os_remove), patch(
        "anyscale.authenticate.AuthenticationBlock.__init__",
        mock_authentication_block_init,
    ):
        auth_controller = AuthController(path="/home/usr0/.anyscale/credentials.json")
        auth_controller.remove()

    mock_os_remove.assert_called_once_with("/home/usr0/.anyscale/credentials.json")
    mock_authentication_block_init.assert_not_called()


def test_auth_show_get_user_info_fails():
    """
    Test auth show but get_user_info is unable to fetch user data.
    This means that the token is invalid or does not exist.
    An exception should occur.
    """
    mock_get_user_info = Mock(return_value=None)
    Mock()
    with patch(
        "anyscale.controllers.auth_controller.get_user_info", mock_get_user_info
    ):
        auth_controller = AuthController()
        auth_controller.show()


def test_auth_show():
    """
    Test auth show command. Must not raise exception.
    """
    userinfo = UserInfo(
        name="Anyscale Test",
        email="anyscale@anyscale.com",
        id="usr_somerandomid",
        organization_permission_level="collaborator",
        organizations=[
            Organization(
                name="Anyscale Inc.",
                id="org_someorgid",
                local_vars_configuration=Mock(client_side_validation=False),
            )
        ],
        local_vars_configuration=Mock(client_side_validation=False),
    )
    mock_get_user_info = Mock(return_value=userinfo)

    with patch(
        "anyscale.controllers.auth_controller.get_user_info", mock_get_user_info
    ):
        auth_controller = AuthController()
        auth_controller.show()



import pytest

from anyscale.utils.cluster_debug import (
    _dns_over_https_test,
    _dns_test,
    _https_test,
    _tls_test,
)


def test_dns_test():
    assert _dns_test("localhost", "127.0.0.1") is None

    mismatched_ip = _dns_test("localhost", "100.100.100.100")
    assert mismatched_ip is not None
    assert "VPN" in mismatched_ip


def test_dns_over_https_test():
    assert _dns_over_https_test("dns.google.com", "8.8.8.8") is None

    mismatched_ip = _dns_over_https_test("dns.google.com", "100.100.100.100")
    assert mismatched_ip is not None
    assert "This should never happen" in mismatched_ip


def test_tls_test():
    assert _tls_test("checkip.amazonaws.com") is None


def test_https_test():
    with pytest.raises(RuntimeError) as exc:
        _https_test("checkip.amazonaws.com")

    exc.match("Unexpected status code")



from dataclasses import dataclass
import socket
import ssl
from typing import Callable, Dict, Iterator, Optional

import requests

from anyscale.cli_logger import BlockLogger


@dataclass
class DebugTestResult:
    name: str
    success: bool
    error: Optional[Exception] = None
    warnings: Optional[str] = None


def debug_cluster(
    logger: BlockLogger, cluster_domain_name: str, anyscale_ip: str
) -> Iterator[DebugTestResult]:
    has_experienced_a_failure: bool = False

    tests = list(_tests_for(cluster_domain_name, anyscale_ip).items())

    for name, test in tests:
        try:
            warning = test()

            yield DebugTestResult(name, True, warnings=warning)
        except Exception as e:  # noqa: BLE001
            yield DebugTestResult(name, False, e)

            if not has_experienced_a_failure:
                has_experienced_a_failure = True

                try:
                    cluster_ip_addr = socket.gethostbyname(cluster_domain_name)
                except Exception as e:  # noqa: BLE001
                    resp = requests.get(
                        "https://dns.google/resolve",
                        params={"name": cluster_domain_name},
                    )
                    resp.raise_for_status()

                    resp_json = resp.json()
                    cluster_ip_addr = resp_json["Answer"][0]["data"]
                logger.warning(
                    "Failure detected, retrying tests by directly communicating with the IP address."
                )
                tests.extend(
                    _tests_for(
                        cluster_ip_addr, anyscale_ip, tls_sni=cluster_domain_name
                    ).items()
                )


def _tests_for(
    endpoint: str, anyscale_ip: str, tls_sni: Optional[str] = None
) -> Dict[str, Callable[[], Optional[str]]]:
    """Attempt to debug networking for an Anyscale Cluster.

    Arguments:
        endpoint: Hostname or IP address of the cluster
        tls_sni: Hostname to use for TLS SNI (used if used to directly dial the IP address)

    Also ask for the IP from the control Plane

    Fallbacks:
    1. Try directly using different IPs + hostname

    """
    return {
        **(
            {
                "dns-lookup": lambda: _dns_test(endpoint, anyscale_ip),
                "dns-over-https": lambda: _dns_over_https_test(endpoint, anyscale_ip),
            }
            if tls_sni is None
            else {}
        ),
        "tcp-connect": lambda: _tcp_test(endpoint),
        "tls-connect": lambda: _tls_test(endpoint, tls_sni),
        "https-get": lambda: _https_test(endpoint, tls_sni),
    }


def _dns_test(endpoint: str, anyscale_ip: str) -> Optional[str]:
    resolved_ip = socket.gethostbyname(endpoint)
    if resolved_ip != anyscale_ip:
        return (
            f"Unexpected local DNS resolution:\nLocal DNS resolves this to {resolved_ip}\n"
            f"Anyscale expects this host to resolve to {anyscale_ip}\n"
            "This is safe to ignore if you are using a VPN to connect to your cluster.\n"
        )
    return None


def _dns_over_https_test(endpoint: str, anyscale_ip: str) -> Optional[str]:
    resp = requests.get("https://dns.google/resolve", params={"name": endpoint})
    resp.raise_for_status()

    resp_json = resp.json()
    answers = resp_json.get("Answer", [])
    assert len(answers) > 0, "Got no answers for a successful query"

    resolved_ips = {answer["data"] for answer in answers}
    if anyscale_ip not in resolved_ips:
        return (
            f"Unexpected Public DNS resolution. This should never happen\n"
            f"`8.8.8.8` resolves this host to {resolved_ips}\n"
            f"Anyscale expects this host to resolve to {anyscale_ip}\n"
        )
    return None


def _tcp_test(endpoint: str) -> None:
    with socket.create_connection((endpoint, 443)):
        pass


def _tls_test(endpoint: str, hostname: Optional[str] = None) -> None:
    if hostname is None:
        hostname = endpoint

    with socket.create_connection(
        (endpoint, 443)
    ) as sock, ssl.create_default_context().wrap_socket(sock, server_hostname=hostname):
        pass


def _https_test(endpoint: str, hostname: Optional[str] = None) -> None:
    if hostname is None:
        hostname = endpoint

    session = requests.Session()
    session.mount(f"https://{endpoint}", HostnameSpecificAdapter(hostname))
    resp = session.get(f"https://{endpoint}")

    if resp.status_code != requests.codes["unauthorized"]:
        raise RuntimeError(
            f"Unexpected status code from https://{endpoint}: {resp.status_code}"
        )

    assert (
        "x-anyscale-authentication-status" in resp.headers
        and resp.headers["x-anyscale-authentication-status"] == "Unauthorized"
    ), "Failed to get an expected anyscale-unique header from the response."


class HostnameSpecificAdapter(requests.adapters.HTTPAdapter):
    """Adapter to handle directly connecting to an IP address with a specific TLS Hostname.

    This allows requests to be made to https://<ip address>, and still have TLS work for the server name.
    """

    def __init__(self, hostname: str, *args, **kwargs):
        super().__init__(*args, **kwargs)

        class SingleHostSSLContext(ssl.SSLContext):
            def wrap_socket(self, *args, **kwargs):
                kwargs["server_hostname"] = hostname
                return super().wrap_socket(*args, **kwargs)

        self.init_poolmanager(
            self._pool_connections,  # type:ignore
            self._pool_maxsize,  # type:ignore
            ssl_context=SingleHostSSLContext(),
            assert_hostname=hostname,
        )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class NodesOptions(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'provider_config': 'object',
        'cluster_name': 'str',
        'instance_ids': 'list[str]',
        'bypass_cold_pool': 'bool'
    }

    attribute_map = {
        'provider_config': 'provider_config',
        'cluster_name': 'cluster_name',
        'instance_ids': 'instance_ids',
        'bypass_cold_pool': 'bypass_cold_pool'
    }

    def __init__(self, provider_config=None, cluster_name=None, instance_ids=None, bypass_cold_pool=False, local_vars_configuration=None):  # noqa: E501
        """NodesOptions - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._provider_config = None
        self._cluster_name = None
        self._instance_ids = None
        self._bypass_cold_pool = None
        self.discriminator = None

        self.provider_config = provider_config
        self.cluster_name = cluster_name
        self.instance_ids = instance_ids
        if bypass_cold_pool is not None:
            self.bypass_cold_pool = bypass_cold_pool

    @property
    def provider_config(self):
        """Gets the provider_config of this NodesOptions.  # noqa: E501


        :return: The provider_config of this NodesOptions.  # noqa: E501
        :rtype: object
        """
        return self._provider_config

    @provider_config.setter
    def provider_config(self, provider_config):
        """Sets the provider_config of this NodesOptions.


        :param provider_config: The provider_config of this NodesOptions.  # noqa: E501
        :type: object
        """
        if self.local_vars_configuration.client_side_validation and provider_config is None:  # noqa: E501
            raise ValueError("Invalid value for `provider_config`, must not be `None`")  # noqa: E501

        self._provider_config = provider_config

    @property
    def cluster_name(self):
        """Gets the cluster_name of this NodesOptions.  # noqa: E501


        :return: The cluster_name of this NodesOptions.  # noqa: E501
        :rtype: str
        """
        return self._cluster_name

    @cluster_name.setter
    def cluster_name(self, cluster_name):
        """Sets the cluster_name of this NodesOptions.


        :param cluster_name: The cluster_name of this NodesOptions.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cluster_name is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_name`, must not be `None`")  # noqa: E501

        self._cluster_name = cluster_name

    @property
    def instance_ids(self):
        """Gets the instance_ids of this NodesOptions.  # noqa: E501


        :return: The instance_ids of this NodesOptions.  # noqa: E501
        :rtype: list[str]
        """
        return self._instance_ids

    @instance_ids.setter
    def instance_ids(self, instance_ids):
        """Sets the instance_ids of this NodesOptions.


        :param instance_ids: The instance_ids of this NodesOptions.  # noqa: E501
        :type: list[str]
        """
        if self.local_vars_configuration.client_side_validation and instance_ids is None:  # noqa: E501
            raise ValueError("Invalid value for `instance_ids`, must not be `None`")  # noqa: E501

        self._instance_ids = instance_ids

    @property
    def bypass_cold_pool(self):
        """Gets the bypass_cold_pool of this NodesOptions.  # noqa: E501


        :return: The bypass_cold_pool of this NodesOptions.  # noqa: E501
        :rtype: bool
        """
        return self._bypass_cold_pool

    @bypass_cold_pool.setter
    def bypass_cold_pool(self, bypass_cold_pool):
        """Sets the bypass_cold_pool of this NodesOptions.


        :param bypass_cold_pool: The bypass_cold_pool of this NodesOptions.  # noqa: E501
        :type: bool
        """

        self._bypass_cold_pool = bypass_cold_pool

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, NodesOptions):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, NodesOptions):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class UserInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'email': 'str',
        'name': 'str',
        'lastname': 'str',
        'username': 'str',
        'verified': 'bool',
        'organization_permission_level': 'OrganizationPermissionLevel',
        'organization_ids': 'list[str]',
        'organizations': 'list[Organization]',
        'has_setup_cloud': 'bool',
        'ld_hash': 'str',
        'ld_hash_fields': 'list[str]',
        'is_support_user': 'bool'
    }

    attribute_map = {
        'id': 'id',
        'email': 'email',
        'name': 'name',
        'lastname': 'lastname',
        'username': 'username',
        'verified': 'verified',
        'organization_permission_level': 'organization_permission_level',
        'organization_ids': 'organization_ids',
        'organizations': 'organizations',
        'has_setup_cloud': 'has_setup_cloud',
        'ld_hash': 'ld_hash',
        'ld_hash_fields': 'ld_hash_fields',
        'is_support_user': 'is_support_user'
    }

    def __init__(self, id=None, email=None, name=None, lastname=None, username=None, verified=None, organization_permission_level=None, organization_ids=None, organizations=None, has_setup_cloud=None, ld_hash=None, ld_hash_fields=None, is_support_user=False, local_vars_configuration=None):  # noqa: E501
        """UserInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._email = None
        self._name = None
        self._lastname = None
        self._username = None
        self._verified = None
        self._organization_permission_level = None
        self._organization_ids = None
        self._organizations = None
        self._has_setup_cloud = None
        self._ld_hash = None
        self._ld_hash_fields = None
        self._is_support_user = None
        self.discriminator = None

        self.id = id
        self.email = email
        self.name = name
        if lastname is not None:
            self.lastname = lastname
        self.username = username
        self.verified = verified
        self.organization_permission_level = organization_permission_level
        self.organization_ids = organization_ids
        self.organizations = organizations
        if has_setup_cloud is not None:
            self.has_setup_cloud = has_setup_cloud
        self.ld_hash = ld_hash
        self.ld_hash_fields = ld_hash_fields
        if is_support_user is not None:
            self.is_support_user = is_support_user

    @property
    def id(self):
        """Gets the id of this UserInfo.  # noqa: E501


        :return: The id of this UserInfo.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this UserInfo.


        :param id: The id of this UserInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def email(self):
        """Gets the email of this UserInfo.  # noqa: E501


        :return: The email of this UserInfo.  # noqa: E501
        :rtype: str
        """
        return self._email

    @email.setter
    def email(self, email):
        """Sets the email of this UserInfo.


        :param email: The email of this UserInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and email is None:  # noqa: E501
            raise ValueError("Invalid value for `email`, must not be `None`")  # noqa: E501

        self._email = email

    @property
    def name(self):
        """Gets the name of this UserInfo.  # noqa: E501


        :return: The name of this UserInfo.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this UserInfo.


        :param name: The name of this UserInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def lastname(self):
        """Gets the lastname of this UserInfo.  # noqa: E501

        The user's last name.  # noqa: E501

        :return: The lastname of this UserInfo.  # noqa: E501
        :rtype: str
        """
        return self._lastname

    @lastname.setter
    def lastname(self, lastname):
        """Sets the lastname of this UserInfo.

        The user's last name.  # noqa: E501

        :param lastname: The lastname of this UserInfo.  # noqa: E501
        :type: str
        """

        self._lastname = lastname

    @property
    def username(self):
        """Gets the username of this UserInfo.  # noqa: E501


        :return: The username of this UserInfo.  # noqa: E501
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username):
        """Sets the username of this UserInfo.


        :param username: The username of this UserInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and username is None:  # noqa: E501
            raise ValueError("Invalid value for `username`, must not be `None`")  # noqa: E501

        self._username = username

    @property
    def verified(self):
        """Gets the verified of this UserInfo.  # noqa: E501


        :return: The verified of this UserInfo.  # noqa: E501
        :rtype: bool
        """
        return self._verified

    @verified.setter
    def verified(self, verified):
        """Sets the verified of this UserInfo.


        :param verified: The verified of this UserInfo.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and verified is None:  # noqa: E501
            raise ValueError("Invalid value for `verified`, must not be `None`")  # noqa: E501

        self._verified = verified

    @property
    def organization_permission_level(self):
        """Gets the organization_permission_level of this UserInfo.  # noqa: E501

        User's permission level in the organization. This value is absent if the user does not have a permission level assigned.  # noqa: E501

        :return: The organization_permission_level of this UserInfo.  # noqa: E501
        :rtype: OrganizationPermissionLevel
        """
        return self._organization_permission_level

    @organization_permission_level.setter
    def organization_permission_level(self, organization_permission_level):
        """Sets the organization_permission_level of this UserInfo.

        User's permission level in the organization. This value is absent if the user does not have a permission level assigned.  # noqa: E501

        :param organization_permission_level: The organization_permission_level of this UserInfo.  # noqa: E501
        :type: OrganizationPermissionLevel
        """
        if self.local_vars_configuration.client_side_validation and organization_permission_level is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_permission_level`, must not be `None`")  # noqa: E501

        self._organization_permission_level = organization_permission_level

    @property
    def organization_ids(self):
        """Gets the organization_ids of this UserInfo.  # noqa: E501

        Deprecated: use organization -- List of organizations that the logged in user is a part of.  # noqa: E501

        :return: The organization_ids of this UserInfo.  # noqa: E501
        :rtype: list[str]
        """
        return self._organization_ids

    @organization_ids.setter
    def organization_ids(self, organization_ids):
        """Sets the organization_ids of this UserInfo.

        Deprecated: use organization -- List of organizations that the logged in user is a part of.  # noqa: E501

        :param organization_ids: The organization_ids of this UserInfo.  # noqa: E501
        :type: list[str]
        """
        if self.local_vars_configuration.client_side_validation and organization_ids is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_ids`, must not be `None`")  # noqa: E501

        self._organization_ids = organization_ids

    @property
    def organizations(self):
        """Gets the organizations of this UserInfo.  # noqa: E501

        List of organizations that the logged in user is a part of.  # noqa: E501

        :return: The organizations of this UserInfo.  # noqa: E501
        :rtype: list[Organization]
        """
        return self._organizations

    @organizations.setter
    def organizations(self, organizations):
        """Sets the organizations of this UserInfo.

        List of organizations that the logged in user is a part of.  # noqa: E501

        :param organizations: The organizations of this UserInfo.  # noqa: E501
        :type: list[Organization]
        """
        if self.local_vars_configuration.client_side_validation and organizations is None:  # noqa: E501
            raise ValueError("Invalid value for `organizations`, must not be `None`")  # noqa: E501

        self._organizations = organizations

    @property
    def has_setup_cloud(self):
        """Gets the has_setup_cloud of this UserInfo.  # noqa: E501

        True if the user has clouds setup in their organization.  # noqa: E501

        :return: The has_setup_cloud of this UserInfo.  # noqa: E501
        :rtype: bool
        """
        return self._has_setup_cloud

    @has_setup_cloud.setter
    def has_setup_cloud(self, has_setup_cloud):
        """Sets the has_setup_cloud of this UserInfo.

        True if the user has clouds setup in their organization.  # noqa: E501

        :param has_setup_cloud: The has_setup_cloud of this UserInfo.  # noqa: E501
        :type: bool
        """

        self._has_setup_cloud = has_setup_cloud

    @property
    def ld_hash(self):
        """Gets the ld_hash of this UserInfo.  # noqa: E501

        Server generated secure hash of the user info that should be sent to LaunchDarkly along with the user data.  # noqa: E501

        :return: The ld_hash of this UserInfo.  # noqa: E501
        :rtype: str
        """
        return self._ld_hash

    @ld_hash.setter
    def ld_hash(self, ld_hash):
        """Sets the ld_hash of this UserInfo.

        Server generated secure hash of the user info that should be sent to LaunchDarkly along with the user data.  # noqa: E501

        :param ld_hash: The ld_hash of this UserInfo.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and ld_hash is None:  # noqa: E501
            raise ValueError("Invalid value for `ld_hash`, must not be `None`")  # noqa: E501

        self._ld_hash = ld_hash

    @property
    def ld_hash_fields(self):
        """Gets the ld_hash_fields of this UserInfo.  # noqa: E501

        List of fields in the userInfo used to generate the secure hash. Clients should send those fields to LaunchDarkly as the user data. These fields should be used in addition to the \"key\" field which is based off the user's id.  # noqa: E501

        :return: The ld_hash_fields of this UserInfo.  # noqa: E501
        :rtype: list[str]
        """
        return self._ld_hash_fields

    @ld_hash_fields.setter
    def ld_hash_fields(self, ld_hash_fields):
        """Sets the ld_hash_fields of this UserInfo.

        List of fields in the userInfo used to generate the secure hash. Clients should send those fields to LaunchDarkly as the user data. These fields should be used in addition to the \"key\" field which is based off the user's id.  # noqa: E501

        :param ld_hash_fields: The ld_hash_fields of this UserInfo.  # noqa: E501
        :type: list[str]
        """
        if self.local_vars_configuration.client_side_validation and ld_hash_fields is None:  # noqa: E501
            raise ValueError("Invalid value for `ld_hash_fields`, must not be `None`")  # noqa: E501

        self._ld_hash_fields = ld_hash_fields

    @property
    def is_support_user(self):
        """Gets the is_support_user of this UserInfo.  # noqa: E501


        :return: The is_support_user of this UserInfo.  # noqa: E501
        :rtype: bool
        """
        return self._is_support_user

    @is_support_user.setter
    def is_support_user(self, is_support_user):
        """Sets the is_support_user of this UserInfo.


        :param is_support_user: The is_support_user of this UserInfo.  # noqa: E501
        :type: bool
        """

        self._is_support_user = is_support_user

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UserInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, UserInfo):
            return True

        return self.to_dict() != other.to_dict()
